;
; This module is a general purpose engine for providing searching and
; completion for tagged function names.
;
; To add support for another language, update tag_case() if it's a case-sensitive
; language, update tags_supported to indicate what file extensions are supported,
; and update proc_search to call the procedure search routine for that language.
;           tag_case()        Returns 'e' for case sensitive languages and
;                            'c' for case insensitive languages.
;
;     xxxxx_proc_search(var proc_name,find_first)
;                             If proc_name is null, this function searches
;                             for a valid procedure in the current buffer. If
;                             successful, proc_name is set to the procedure
;                             name and 0 is returned.  The find_first parameter
;                             when non-zero indicates that the first search
;                             is being performed.
;
;                             If proc_name is NOT null, this function searches
;                             for the definition of the procedure proc_name in
;                             the current buffer.  If successful, cursor is
;                             placed on procedure definition and 0 is returned.
;                             See one of the procedures C_PROC_SEARCH,
;                             PAS_PROC_SEARCH, or ASM_PROC_SEARCH for an
;                             example.


defproc tag_case(filename)
   ext=filetype(filename)
   if ext='c' then   /* Case sensitive language? */
      return 'e'
   endif
   return 'c'  /* Case insensitive language? */

defproc tags_supported(ext)
   return wordpos(ext, 'C E ASM PAS PASCAL')

defproc proc_search(var proc_name, first_flag, ext)
   if ext = 'C' then
      return c_proc_search(proc_name, first_flag)
   elseif ext = 'ASM' then
      return asm_proc_search(proc_name, first_flag)
   elseif ext = 'PAS' | ext = 'PASCAL' then
      return pas_proc_search(proc_name, first_flag)
   elseif ext = 'E' then
      return e_proc_search(proc_name, first_flag)
   else
      return 1
   endif

/****   The above is all that needs to be modified for adding other languages. *****/

compile if not defined(SMALL)  -- If SMALL not defined, then being separately
 define INCLUDING_FILE = 'TAGS.E'
const
   tryinclude 'MYCNF.E'        -- the user's configuration customizations.

 compile if not defined(SITE_CONFIG)
    const SITE_CONFIG = 'SITECNF.E'
 compile endif
 compile if SITE_CONFIG
    tryinclude SITE_CONFIG
 compile endif

include 'stdconst.e'

const
 compile if not defined(NLS_LANGUAGE)
   NLS_LANGUAGE = 'ENGLISH'
 compile endif
include NLS_LANGUAGE'.e'
compile endif

const
compile if not defined(TAGS_ANYWHERE)
   TAGS_ANYWHERE = 1          -- Set to 0 if all your procedure definitions start in col. 1
compile endif
compile if not defined(KEEP_TAGS_FILE_LOADED)
   KEEP_TAGS_FILE_LOADED = 1  -- If you do a lot with tags, you might want to keep the file loaded.
compile endif
   IDENTIFIER_STARTER = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ_$'

defc tagsfile
   universal tags_file
compile if KEEP_TAGS_FILE_LOADED
   universal tags_fileid
compile endif

   orig_name = tags_file
   if arg(1)='' then
      parse value entrybox(TAGSNAME__MSG,'/'SET__MSG'/'SETP__MSG'/'Cancel__MSG'/'Help__MSG'/',tags_filename(),'',200,
             atoi(1) || atoi(6070) || gethwndc(APP_HANDLE) ||
             TAGSNAME_PROMPT__MSG) with button 2 newname \0
      if button=\1 | button=\2 then
         tags_file = newname
         if button=\2 & tags_file<>'' then
            call setini('TAGSFILE', tags_file)
         endif
      endif
   else
      tags_file = arg(1)
   endif
compile if KEEP_TAGS_FILE_LOADED
   if tags_fileid <> '' & orig_name <> tags_file then  -- New name; drop tags file
      getfileid startfid
      rc = 0
      activatefile tags_fileid
      if rc=0 then 'quit'; endif
      activatefile startfid
   endif
compile endif


defproc tags_filename()
   universal tags_file
   if tags_file='' then
      tags_file=checkini(0, 'TAGSFILE', '')
   endif
   if tags_file='' then
      tags_file=get_env('TAGS.EPM')
   endif
   if tags_file='' then
      tags_file='tags.epm'
   endif
   return(tags_file)

defc find_tag, findtag
compile if KEEP_TAGS_FILE_LOADED
   universal tags_fileid
compile endif
   button = ''
   if arg(1)='' then
      /* Try to find the procedure at the cursor. */
      if not find_token(startcol, endcol) then
         return 1
      endif
      proc_name = substr(textline(.line), startcol, (endcol-startcol)+1)
   elseif arg(1)='*' then
      parse value entrybox(FINDTAG__MSG,'/'OK__MSG'/'LIST__MSG'/'Cancel__MSG'/'Help__MSG'/',\0,'',200,
             atoi(1) || atoi(6010) || gethwndc(APP_HANDLE) ||
             FINDTAG_PROMPT__MSG) with button 2 proc_name \0
      if button<>\1 & button<>\2 then return; endif
   else
      proc_name = arg(1)
   endif
compile if KEEP_TAGS_FILE_LOADED
   getfileid startfid
   if tags_fileid<>'' then
      rc = 0
      display -2
      activatefile tags_fileid
      display 2
      if rc then
         tags_fileid = ''
      else
         0              -- Go to top of file
      endif
   endif
   if tags_fileid='' then
compile endif
      'e /d ' tags_filename()
      if rc then
         if rc=-282 then  -- -282 = sayerror("New file")
            'quit'
            sayerror "Tag file '"tags_filename()"' not found"
         else
            sayerror "Error loading tag file '"tags_filename()"' -" sayerrortext(rc)
         endif
         return 1
      endif
compile if KEEP_TAGS_FILE_LOADED
      .visible = 0
      getfileid tags_fileid
   endif
compile endif
   if button=\2 then  -- List (delayed until tags_file was loaded)
      sayerror BUILDING_LIST__MSG
      'xcom e /c tagslist'
      if rc<>-282 then  -- -282 = sayerror("New file")
         sayerror ERROR__MSG rc BAD_TMP_FILE__MSG sayerrortext(rc)
         return
      endif
      browse_mode = browse()     -- query current state
      if browse_mode then call browse(0); endif
      .autosave = 0
      getfileid lb_fid
      display -2
      do i=1 to tags_fileid.last
         getline line, i, tags_fileid
         parse value line with tag .
         if tag<>'' & tag<>'*' then
            insertline tag, .last+1
         endif
      enddo
      if browse_mode then call browse(1); endif  -- restore browse state
      display 2
      if not .modify then  -- Nothing added?
         'xcom quit'
compile if KEEP_TAGS_FILE_LOADED
         activatefile startfid
compile else
         'quit'
compile endif
         sayerror NO_TAGS__MSG
         return
      endif
      buflen = filesize() + .last + 1
;sayerror 'buflen=' buflen'; tags file size =' tags_fileid.last'; # listbox entries =' .last'; last entry =' textline(.last)
      if buflen > MAXBUFSIZE then
         sayerror LIST_TOO_BIG__MSG '(' buflen '>' MAXBUFSIZE ')'
         buflen = MAXBUFSIZE
      endif
      bufhndl = buffer(CREATEBUF, 'LISTBOX', buflen, 1 )  -- create a private buffe
      if not bufhndl then sayerror 'CREATEBUF' ERROR_NUMBER__MSG RC; return; endif
      noflines = buffer(PUTBUF, bufhndl, 1, 0, APPENDCR)
      .modify = 0
      'xcom quit'
      if not noflines then sayerror 'PUTBUF' ERROR_NUMBER__MSG RC; return; endif
      usedsize = buffer(USEDSIZEBUF,bufhndl)
      parse value listbox(LIST_TAGS__MSG,
    compile if EPM32
                          \0 || atol(usedsize) || atoi(32) || atoi(bufhndl),
    compile else
                          \0 || atoi(usedsize) || atoi(bufhndl) || atoi(32),
    compile endif -- EPM32
                          '/'OK__MSG'/'Cancel__MSG'/'Help__MSG,1,5,min(noflines,12),0,
    compile if EVERSION >= 5.60
                          gethwndc(APP_HANDLE) || atoi(1) || atoi(1) || atoi(6012)) with button 2 proc_name \0
    compile else
                          atoi(1) || atoi(1) || atoi(6012) || gethwndc(APP_HANDLE)) with button 2 proc_name \0
    compile endif
      call buffer(FREEBUF, bufhndl)
      if button<>\1 then
compile if KEEP_TAGS_FILE_LOADED
         activatefile startfid
compile else
         'quit'
compile endif
         return
      endif
      activatefile tags_fileid
   endif
   name = proc_name  -- Preserve original name.
   display -2
   do i=1 to 2
      'l ^'proc_name' x'tag_case(filename)
      if not rc then leave; endif
      proc_name = '_'proc_name  /* Handle case where C call to assembler function needs '_' */
   enddo
   display 2
   long_msg='.  You may want to rebuild the tag file.'
   if rc then
compile if KEEP_TAGS_FILE_LOADED
      activatefile startfid
compile else
      'quit'
compile endif
      sayerror 'Tag for function "'name'" not found in 'tags_filename()long_msg
      return 1
   endif
   parse_tagline(name, filename, fileline, filedate)
   /* Check if there is more than one */
   if .line < .last then
      '+1'
      parse_tagline(next_name, next_filename, next_fileline, next_filedate)
      if upcase(name)=upcase(next_name) then
         getfileid tags_fid
         'xcom e /c temp'
         if rc<>-282 then  -- -282 = sayerror("New file")
            'xcom quit'
            return 1
         endif
         browse_mode = browse()     -- query current state
         if browse_mode then call browse(0); endif
         getfileid temp_fid
         .autosave = 0
         insertline filename, 2
         activatefile tags_fid
         loop
            if upcase(next_filename) <> upcase(filename) then
               insertline next_filename, temp_fid.last+1, temp_fid
            endif
            if .line = .last then
               leave
            endif
            '+1'
            parse_tagline(next_name, next_filename, next_fileline, next_filedate)
            if upcase(name)/==upcase(next_name) then
               leave
            endif
         endloop
         activatefile temp_fid
         .modify = 0
         if browse_mode then call browse(1); endif  -- restore browse state
         if .last>2 then
            buflen = filesize() + .last + 1
            bufhndl = buffer(CREATEBUF, 'LISTBOX', buflen, 1 )  -- create a private buffer
            if not bufhndl then sayerror 'CREATEBUF' ERROR_NUMBER__MSG RC; return; endif
            noflines = buffer(PUTBUF, bufhndl, 1, 0, APPENDCR)
            'xcom quit'
            if not noflines then sayerror 'PUTBUF' ERROR_NUMBER__MSG RC; return; endif
            usedsize = buffer(USEDSIZEBUF,bufhndl)
            parse value listbox('Select a file',
          compile if EPM32
                                \0 || atol(usedsize) || atoi(32) || atoi(bufhndl),
          compile else
                                \0 || atoi(usedsize) || atoi(bufhndl) || atoi(32),
          compile endif -- EPM32
                                '/'OK__MSG'/'Cancel__MSG'/'Help__MSG,0,0,min(noflines,12),60,
          compile if EVERSION >= 5.60
                                gethwndc(APP_HANDLE) || atoi(1) || atoi(1) || atoi(6015) ||
          compile else
                                atoi(1) || atoi(1) || atoi(6015) || gethwndc(APP_HANDLE) ||
          compile endif
                                "'"name"' appears in multiple files.") with button 2 filename \0
            call buffer(FREEBUF, bufhndl)
            if button<>\1 then  -- Didn't select OK
               filename = ''
            endif
            fileline = ''; filedate = ''  -- For now, don't try to keep track.
         else
            'xcom quit'
         endif
         if filename='' then
compile if KEEP_TAGS_FILE_LOADED
            activatefile startfid
compile else
            'quit'  -- quit tags file
compile endif
            return 1
         endif
      endif  -- duplicate names
   endif  -- not on last line
compile if KEEP_TAGS_FILE_LOADED
   activatefile startfid
compile else
   'quit'  -- quit tags file
compile endif

   getfileid already_loaded, filename
   'e /v' filename
   if rc then
      if rc=-282 then  -- -282 = sayerror("New file")
         'q'
         sayerror "'"filename"' not found"long_msg
      else
         sayerror "Error loading '"filename"' -" sayerrortext(rc)
      endif
      return 1
   endif
   if filedate<>''  then  -- Line number and file write date preserved
      if filedate=get_file_date(filename) then  -- Same date means file has not been changed,
         display -8
         sayerror 'Jumping straight to line.'  /**/
         display 8
         fileline                               -- so we can jump right to the line.
         return
      endif
   endif
   display -8
   sayerror 'Searching for routine.'  /**/
   display 8
   top
   rc=proc_search(proc_name,1,filetype())
   if rc then
      if already_loaded='' then 'quit' endif
      sayerror proc_name" not found in '"filename"'"long_msg
      return 1
   endif
   if already_loaded<>'' then
      sayerror 'File already loaded, starting new view.'
   endif

defproc parse_tagline(var name, var filename, var fileline, var filedate)
   parse value textline(.line) with name filename fileline filedate .
   if leftstr(filename,1)='"' & (rightstr(filename,1)<>'"' | length(filename)=1) then
      parse value textline(.line) with name ' "'filename'"' fileline filedate .
      filename = '"'filename'"'
   endif

const
   IGNORE_C_KEYWORDS = 'if while switch for case else return ='
compile if not defined(LOG_TAG_MATCHES)
   LOG_TAG_MATCHES = 0
compile endif

defproc c_proc_search(var proc_name, find_first)
compile if LOG_TAG_MATCHES
   universal TAG_LOG_FID
compile endif
   proc_len = length(proc_name)
   display -2
   if find_first then
      if proc_name=='' then
compile if TAGS_ANYWHERE
         'xcom l ^:fex'
compile else
         'xcom l ^[A-Za-z_$].*\(ex'
compile endif
      else
         'xcom l 'proc_name'e'
      endif
   else
      repeat_find
   endif
   loop
      if rc then
         display 2
         return rc
      endif
      getline line
      line=translate(line, ' ', \t)
compile if LOG_TAG_MATCHES
         if TAG_LOG_FID then
            insertline '  Found line' .line '= "'line'"', TAG_LOG_FID.last+1, TAG_LOG_FID
         endif
compile endif
      if proc_len then  -- Determine if match is a substring of something else
         if .col>1 then
            c = upcase(substr(line, .col-1, 1))
            if (c>='A' & c<='Z') | (c>='0' & c<='9') | c='$' | c='_'  then
               end_line; repeat_find; iterate
            endif
         endif
         .col = .col + proc_len
         c = upcase(substr(line, .col, 1))
         if (c>='A' & c<='Z') | (c>='0' & c<='9') | c='$' | c='_'  then
            end_line; repeat_find; iterate
         endif
         p = pos('(', line, .col)
         if not p then
            end_line; repeat_find; iterate
         endif
         .col = p
      else
         .col = pos('(', line)
      endif
      /* Strip trailing comment.  */
      i=pos('//',line)
      if i then
         line=leftstr(line,i-1)
      endif
      loop
         i=pos('/*',line)
         if not i then leave; endif
         j=pos('*/', line, i+2)
         if j then
            /* line=delstr(line,i,j-i+2) */
            line=overlay('', line, i, j-i+2)  -- Keep column alignment
         else
            line=leftstr(line,i-1)
         endif
      endloop
      line = strip(line, 'T')
      if substr(line, .col, 1)='(' & rightstr(line,1)<>';' then
         if rightstr(line,1)<>')' | pos('(',line, .col+1) then
;;          .col=pos('(',line,.col)
            call psave_pos(save_pos)
            if find_matching_paren() then  -- No match found?
compile if LOG_TAG_MATCHES
               if TAG_LOG_FID then
                  insertline '  ...skipping; no matching paren found.', TAG_LOG_FID.last+1, TAG_LOG_FID
               endif
compile endif
               end_line; repeat_find; iterate  -- Keep looking
            endif
            after_paren_ch=leftstr(strip(substr(translate(textline(.line), ' ', \t),.col+1)),1)
            do while after_paren_ch = ' ' & .line<.last
               '+1'
               after_paren_ch=leftstr(strip(translate(textline(.line), ' ', \t)),1)
            enddo
            if pos(after_paren_ch,';),([-+*.=?&|}') then
compile if LOG_TAG_MATCHES
               if TAG_LOG_FID then
                  insertline '  ...skipping; after_paren_ch in list.  "'after_paren_ch'"', TAG_LOG_FID.last+1, TAG_LOG_FID
               endif
compile endif
               end_line; repeat_find; iterate
            endif
            call prestore_pos(save_pos)
         endif
         parse value strip(line) with line '('
         proc_name = lastword(line)
         v = verify(upcase(proc_name), IDENTIFIER_STARTER, 'M')
         if not v then
compile if LOG_TAG_MATCHES
            if TAG_LOG_FID then
               insertline '  ...skipping; verify =' v, TAG_LOG_FID.last+1, TAG_LOG_FID
            endif
compile endif
            end_line; repeat_find; iterate
         endif
         proc_name = substr(proc_name, v)
         if wordpos(proc_name, IGNORE_C_KEYWORDS) then
compile if LOG_TAG_MATCHES
            if TAG_LOG_FID then
               insertline '  ...skipping; procname "'proc_name'" in ignore list', TAG_LOG_FID.last+1, TAG_LOG_FID
            endif
compile endif
            end_line; repeat_find; iterate
         endif
         if verify(upcase(proc_name), IDENTIFIER_STARTER'0123456789') then
compile if LOG_TAG_MATCHES
            if TAG_LOG_FID then
               insertline '  ...skipping; procname "'proc_name'" contains invalid characters', TAG_LOG_FID.last+1, TAG_LOG_FID
            endif
compile endif
            end_line; repeat_find; iterate
         endif
         w=words(line)
         if w>1 then
            if verify(upcase(subword(line,1,w-1)), IDENTIFIER_STARTER'0123456789*()[] ') then
compile if LOG_TAG_MATCHES
               if TAG_LOG_FID then
                  insertline '  ...skipping; character invalid in a declarator appears before "'proc_name'" in:  'line, TAG_LOG_FID.last+1, TAG_LOG_FID
               endif
compile endif
               end_line; repeat_find; iterate
            endif
         endif

         display 2
compile if LOG_TAG_MATCHES
         if TAG_LOG_FID then
            insertline '  ...accepted; proc_name = "'proc_name'"', TAG_LOG_FID.last+1, TAG_LOG_FID
         endif
compile endif
         return 0
compile if LOG_TAG_MATCHES
      elseif not TAG_LOG_FID then  -- do nothing
      elseif substr(line, .col, 1)<>'(' then
         insertline '  ...skipping; .col =' .col'; char = "'substr(line, .col, 1)'"', TAG_LOG_FID.last+1, TAG_LOG_FID
      elseif rightstr(line,1)=';' then
         insertline '  ...skipping; ends in semicolon; line = "'line'"', TAG_LOG_FID.last+1, TAG_LOG_FID
      else
         insertline '  ...skipping; no idea why.', TAG_LOG_FID.last+1, TAG_LOG_FID
compile endif
      endif
      end_line; repeat_find
   endloop



defproc pas_proc_search(var proc_name,find_first)
   tc=arg(3)
   if tc='' then  /* pascal search?*/
      tc='c'  /* ignore case */
   endif
   proc_len = length(proc_name)
   display -2
   if find_first then
      if tc='e' then  /* Must be modula search */
         keywords='(PROCEDURE)'
      else
         keywords='(overlay:b|)(pro(cedure|gram)|function)'
      endif
      if proc_name=='' then
          'xcom l ^[ \t]*'keywords'[ \t]+[A-Za-z_$][A-Za-z0-9_$]*'||
                 '[( \t;:]xc'
      else
         'xcom l 'proc_name''tc
      endif
   else
      repeat_find
   endif
;  if tc='e' then /* pos function does not support allow e option*/
      tc=''
;  endif
   loop
      if rc then
         display 2
         return(rc)
      endif
      getline line
      if proc_len then  -- Determine if match is a substring of something else
         if .col>1 then
            c = upcase(substr(line, .col-1, 1))
            if (c>='A' & c<='Z') | (c>='0' & c<='9') | c='$' | c='_'  then
               end_line; repeat_find; iterate
            endif
         endif
         .col = .col + proc_len
         c = upcase(substr(line, .col, 1))
         if (c>='A' & c<='Z') | (c>='0' & c<='9') | c='$' | c='_'  then
            end_line; repeat_find; iterate
         endif
      else
         .col = pos('(', line)
      endif
      line=translate(line, ' ', \t)
      col=.col
      if not pos(' 'keywords'[ \t]',' 'line,'','x'/*||tc*/) then
         end_line; repeat_find; iterate
      endif
      p=pos('[(;:]',line,'','x')
      if p then
         if substr(line,p,1)=='(' then
            .col=p
            call psave_pos(save_pos)
            if find_matching_paren() then
               end_line; repeat_find; iterate
            endif
            call prestore_pos(save_pos)
         endif
         if pos('forward;',textline(.line)) then
            end_line; repeat_find; iterate
         endif
         line=substr(line,1,p-1)
         i=lastpos(' ',strip(translate(line,' ',\t)))
         proc_name=substr(line,i+1)
         display 2
         return 0
      endif
      end_line; repeat_find
   endloop

defproc asm_proc_search(var proc_name,find_first)
compile if LOG_TAG_MATCHES
   universal TAG_LOG_FID
compile endif
   display -2
   if find_first then
      if proc_name=='' then
          proc_name=':c'
      endif
      'xcom l ^'proc_name':wproc(:w|$)xc'
   else
      repeat_find
   endif
   display 2
   parse value translate(textline(.line), ' ', \t) with proc_name .
compile if LOG_TAG_MATCHES
   if TAG_LOG_FID and not rc then
      insertline '  Found proc_name = "'proc_name'" in line' .line '= "'textline(.line)'"', TAG_LOG_FID.last+1, TAG_LOG_FID
   endif
compile endif
   return rc

defproc e_proc_search(var proc_name, find_first)
compile if LOG_TAG_MATCHES
   universal TAG_LOG_FID
compile endif
   display -2
   proc_len = length(proc_name)
   if find_first then
      if proc_name=='' then
          proc_name='[A-Z_]'
      endif
compile if TAGS_ANYWHERE
      'xcom l ^:oDEFPROC:w'proc_name'cx'
compile else
      'xcom l ^DEFPROC:w'proc_name'cx'
compile endif
   else
      repeat_find
   endif
   loop
      if rc then
         display 2
         return rc
      endif
      parse value translate(textline(.line), ' ', \t) with . proc_name .
      parse value proc_name with proc_name '('
      if proc_len then
         if length(proc_name)<>proc_len then  -- a substring of something else
            end_line; repeat_find; iterate
         endif
      endif
compile if LOG_TAG_MATCHES
      if TAG_LOG_FID and not rc then
         insertline '  Found proc_name = "'proc_name'" in line' .line '= "'textline(.line)'"', TAG_LOG_FID.last+1, TAG_LOG_FID
      endif
compile endif
      leave
   endloop
   display 2
   return rc

defc make_tags
   'maketags' arg(1)

defproc find_matching_paren
   n=1
   GETSEARCH search_command -- Save user's search command.
   display -2
   'xcom L /[\(\)]/ex+F'
   loop
      repeatfind
      if rc then leave; endif
      if substr(textline(.line), .col, 1) = '(' then n=n+1; else n=n-1; endif
      if n=0 then leave; endif
   endloop
   display 2
   SETSEARCH search_command -- Restores user's command so Ctrl-F works.
   return rc  /* 0 if found, else sayerror('String not found') */

defproc get_file_date(filename)
         pathname = filename\0
         resultbuf = copies(\0,30)
compile if EVERSION >= 6  -- EPM32:  32-bit version
         ca = dynalink32('DOSCALLS',      /* dynamic link library name       */
                         '#223',           /* ordinal value for DOS32QueryPathInfo  */
                         address(pathname)         ||  -- pathname to be queried
                         atol(1)                   ||  -- PathInfoLevel
                         address(resultbuf)        ||  -- buffer where info is to be returned
                         atol(length(resultbuf)) )     -- size of buffer
compile else
         ca = dynalink('DOSCALLS',      /* dynamic link library name       */
                       '#98',           /* ordinal value for DOSQPATHINFO  */
                       address(pathname)         ||  -- pathname to be queried
                       atoi(1)                   ||  -- PathInfoLevel
                       address(resultbuf)        ||  -- buffer where info is to be returned
                       atoi(length(resultbuf))   ||  -- size of buffer
                       atol(0) )                     -- (reserved)
compile endif -- EVERSION >= 6
         return ltoa(substr(resultbuf, 9, 4), 16)
